class Boid {

  Vector3D loc;
  Vector3D vel;
  Vector3D acc;
  float r;
  float maxforce;    // Maximum steering force
  float maxspeed = speed;    // Maximum speed
  
  float neighbordist;
  float desiredseparation;
  float perceptionangle;
  ArrayList neighbors;
  
  Subject subject;
  
  // for visualization
  boolean neighborhoodrepresentation;
  boolean connectivity;
  
  Boid(Vector3D l, float ms, float mf, Subject s) {
    acc = new Vector3D(0,0);
    vel = new Vector3D(random(-1,1),random(-1,1));
    loc = l.copy();
    r = 2.0f;
    maxspeed = ms;
    maxforce = mf;
    subject = s;
    
    neighbors = new ArrayList();
    neighbordist = 60f;
    perceptionangle = 1f;
    desiredseparation = 15f;
    neighborhoodrepresentation = false;
    connectivity = false;
  }
  
  void run(ArrayList boids) {
    chaseDeweyAttractor(matrix.deweyAttractor);
    updateNeighbors(boids);
    flock(neighbors);
    chaseDeweyAttractor(matrix.deweyAttractor);
    
    if (pause == false) 
    {
      updateFlocking();
    }

    borders();
    render();
  }

  // We accumulate a new acceleration each time based on three rules
  void flock(ArrayList boids) {
    Vector3D sep = separate(boids);   // Separation
    Vector3D ali = align(boids);      // Alignment
    Vector3D coh = cohesion(boids);   // Cohesion
    // Arbitrarily weight these forces
    sep.mult(2.0f);
    ali.mult(1.6f);
    coh.mult(1.6f);
    // Add the force vectors to acceleration
    acc.add(sep);
    acc.add(ali);
    acc.add(coh);
    
    Vector3D rand = new Vector3D(random(-1,1),random(-1,1),random(-1,1));
    rand.mult(0.1f);
    acc.add(rand);
        
  }
  
  // Method to update location
  void updateFlocking() {
    // Update velocity
    vel.add(acc);
    // Limit speed
    vel.limit(maxspeed);
    loc.add(vel);
    
    // avoid flying into a particle
    for (int i = 0 ; i < matrix.deweyAttractor.size(); i++) { 
      DeweyAttractor p = (DeweyAttractor) matrix.deweyAttractor.get(i);
      if (loc.distance(loc, p.loc) <= p.radius) loc.sub(vel);
    }
    // Reset accelertion to 0 each cycle
    acc.setXYZ(0,0,0);
  }

  void updateNeighbors(ArrayList boids) {
    // forget your former neighbors
    neighbors.clear();
    for (int i = 0 ; i < boids.size(); i++) 
    {   
      Boid other = (Boid) boids.get(i);
      Vector3D d = new Vector3D();
      d = d.sub(loc, other.loc);
      // If the distance is greater than 0 and less than an arbitrary amount (0 when you are yourself)
      if ((d.length() > 0) && (d.length() < neighbordist))
      {
        if (vel.angleBetweenTwo2D(d) <= perceptionangle / 2.0)
        {
          neighbors.add(other);   
          
          if (connectivity == true) 
          {
            noFill();
            strokeWeight(lineweight/80);
            stroke(90);
            line(loc.x,loc.y,loc.z,other.loc.x,other.loc.y,other.loc.z);   
            strokeWeight(0.2);
          }
        }
      }
    }
  }

  void chaseDeweyAttractor(ArrayList<DeweyAttractor> DeweyAttractor) {
    for (int i = 0 ; i < DeweyAttractor.size(); i++) {   
      if (subject.A_dewey.contains(DeweyAttractor.get(i).deweyDict.id))
      {
        DeweyAttractor other = (DeweyAttractor) DeweyAttractor.get(i);
       
        Vector3D d = new Vector3D();
        d = d.sub(loc, other.loc);
        float effectiveDist = d.length() - other.radius; // if negative, we are within (!) the circle
         
        if( effectiveDist < neighbordist) {
          
          
          // we know that the particle is within the agent's radius
          // but is it also within the angle of its field of perception?
          // (1) the abs(angle) between AGENT | PARTICLE => 0 and <= alpha / 2.0
          // (2) test intersection right-hand border
          // (3) test intersection left-hand border
          
          Vector3D border = vel.copy();
          boolean intersection = false; 
          
          if ( (0 >= abs(vel.angleBetweenTwo2D(other.loc))) && 
               (abs(vel.angleBetweenTwo2D(other.loc)) <= perceptionangle / 2.0) ) intersection = true;
          
          border.rotate2D(perceptionangle /2.0);
          if (border.sphereIntersection(loc, other.loc, other.radius)) intersection = true;
          
          border.rotate2D(- perceptionangle);        
          if (border.sphereIntersection(loc, other.loc, other.radius))  intersection = true;
          
          if (!intersection) continue;
  
          float weight = DeweyAttractForce / (effectiveDist);
          acc.x = vel.x * weight;
          acc.y = vel.y * weight;
          acc.rotate2D(vel.angleBetweenTwo2D(d));
        }
      }
    }
  }

  void seek(Vector3D target) {
    acc.add(steer(target,false));
  }
 
  void arrive(Vector3D target) {
    acc.add(steer(target,true));
  }

  // A method that calculates a steering vector towards a target
  // Takes a second argument, if true, it slows down as it approaches the target
  Vector3D steer(Vector3D target, boolean slowdown) {
    Vector3D steer;  // The steering vector
    Vector3D desired = target.sub(target,loc);  // A vector pointing from the location to the target
    float d = desired.magnitude(); // Distance from the target is the magnitude of the vector
    // If the distance is greater than 0, calc steering (otherwise return zero vector)
    if (d > 0) {
      // Normalize desired
      desired.normalize();
      // Two options for desired vector magnitude (1 -- based on distance, 2 -- maxspeed)
      if ((slowdown) && (d < 100.0f)) desired.mult(maxspeed*(d/100.0f)); // This damping is somewhat arbitrary
      else desired.mult(maxspeed);
      // Steering = Desired minus Velocity
      steer = target.sub(desired,vel);
      steer.limit(maxforce);  // Limit to maximum steering force
    } else {
      steer = new Vector3D(0,0);
    }
    return steer;
  }
  
  void render() {
    if (neighborhoodrepresentation == true)
    {
      noFill();
      stroke(150);
    }
    
    
    pushMatrix();
    strokeWeight(lineweight);
    point(loc.x,loc.y,loc.z);
    popMatrix();
  }
  
  // Wraparound
  void borders() {
    if (loc.x < -r) loc.x = worldSize.x+r;
    if (loc.y < -r) loc.y = worldSize.y+r;
    if (loc.z < -r) loc.z = worldSize.z+r;
    if (loc.x > worldSize.x+r) loc.x = -r;
    if (loc.y > worldSize.y+r) loc.y = -r;
    if (loc.z > worldSize.z+r) loc.z = -r;
 
     pushMatrix();
     noFill();
     translate(worldSize.x/2.0,worldSize.y/2.0,worldSize.z/2.0);

     strokeWeight(0.5);
     stroke(0,0,0);
     box(worldSize.x,worldSize.y,worldSize.z);
     popMatrix();
  }

  // Separation
  // Method checks for nearby boids and steers away
  Vector3D separate (ArrayList neighbors) {
    Vector3D sum = new Vector3D(0,0,0);
    int count = 0;
    // For every boid in the system, check if it's too close
    for (int i = 0 ; i < neighbors.size(); i++) {
      Boid other = (Boid) neighbors.get(i);
      if (!subject.A_subject.contains(other.subject.id))
      {
        float d = loc.distance(loc,other.loc);
  
        // too close? then separate!
        if (d < desiredseparation) {
          // Calculate vector pointing away from neighbor
          Vector3D diff = loc.sub(loc,other.loc);
          diff.normalize();
          diff.div(d);        // Weight by distance
          sum.add(diff);
          count++;            // Keep track of how many
        }
      }
    }
    // Average -- divide by how many
    if (count > 0) {
      sum.div((float)count);
    }
    return sum;
  }
  
  // Alignment
  // For every nearby boid in the system, calculate the average velocity
  Vector3D align (ArrayList neighbors) {
    Vector3D sum = new Vector3D(0,0,0);
    // no neighbors, let's stop this.
    if (neighbors.size() == 0) return sum;
    for (int i = 0 ; i < neighbors.size(); i++) {
      Boid other = (Boid) neighbors.get(i);
      sum.add(other.vel);
      }
    sum.div((float)neighbors.size());
    sum.limit(maxforce);
    return sum;
  }

  // Cohesion
  // For the average location (i.e. center) of all nearby boids, calculate steering vector towards that location
  Vector3D cohesion (ArrayList neighbors) {
    Vector3D sum = new Vector3D(0,0,0);   // Start with empty vector to accumulate all locations
    // no neighbors, let's stop this.
    if (neighbors.size() == 0) return sum;
    for (int i = 0 ; i < neighbors.size(); i++) {
      Boid other = (Boid) neighbors.get(i);
      float d = loc.distance(loc,other.loc);
      sum.add(other.loc); // Add location
    }
    sum.div((float)neighbors.size());
    return steer(sum,false);  // Steer towards the location
  }
}
